# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
from immudb.constants import *
from immudb.exceptions import *


class InclusionProof:
    def __init__(self):
        self.leaf = None
        self.width = None
        self.terms = b''


class HTree:
    def __init__(self, maxWidth: int):
        if maxWidth < 1:
            return

        self.maxWidth = maxWidth
        lw = 1
        while lw < maxWidth:
            lw = lw << 1
        height = (maxWidth-1).bit_length()+1
        self.levels = [None]*height
        for l in range(0, height):
            self.levels[l] = [None]*(lw >> l)

    def BuildWith(self, digests: list):
        if len(digests) > self.maxWidth:
            raise ErrMaxWidthExceeded
        if len(digests) == 0:
            raise ErrIllegalArguments
        for i in range(0, len(digests)):
            leaf = LEAF_PREFIX+digests[i]
            self.levels[0][i] = hashlib.sha256(leaf).digest()
        l = 0
        w = len(digests)
        while w > 1:

            wn = 0
            i = 0
            while i+1 < w:
                b = NODE_PREFIX+self.levels[l][i]+self.levels[l][i+1]
                self.levels[l+1][wn] = hashlib.sha256(b).digest()
                wn = wn+1
                i = i+2
            if w % 2 == 1:
                self.levels[l+1][wn] = self.levels[l][w-1]
                wn = wn+1
            l += 1
            w = wn
        self.width = len(digests)
        self.root = self.levels[l][0]

    def InclusionProof(self, i):
        if i >= self.width:
            raise ErrIllegalArguments
        m = i
        n = self.width
        offset = 0
        proof = InclusionProof()
        proof.leaf = i
        proof.width = self.width
        if self.width == 1:
            return proof
        while True:
            d = (n-1).bit_length()
            k = 1 << (d-1)
            if m < k:
                l, r = offset+k, offset+n-1
                n = k
            else:
                l, r = offset, offset+k-1
                m = m-k
                n = n-k
                offset = offset+k
            layer = (r-l).bit_length()
            index = int(l/(1 << layer))
            proof.terms = self.levels[layer][index]+proof.terms
            if n < 1 or (n == 1 and m == 0):
                return proof



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import schema_pb2 as schema__pb2


class ImmuServiceStub(object):
  """immudb gRPC & REST service
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ListUsers = channel.unary_unary(
        '/immudb.schema.ImmuService/ListUsers',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.UserList.FromString,
        )
    self.CreateUser = channel.unary_unary(
        '/immudb.schema.ImmuService/CreateUser',
        request_serializer=schema__pb2.CreateUserRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.ChangePassword = channel.unary_unary(
        '/immudb.schema.ImmuService/ChangePassword',
        request_serializer=schema__pb2.ChangePasswordRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.ChangePermission = channel.unary_unary(
        '/immudb.schema.ImmuService/ChangePermission',
        request_serializer=schema__pb2.ChangePermissionRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.SetActiveUser = channel.unary_unary(
        '/immudb.schema.ImmuService/SetActiveUser',
        request_serializer=schema__pb2.SetActiveUserRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.UpdateAuthConfig = channel.unary_unary(
        '/immudb.schema.ImmuService/UpdateAuthConfig',
        request_serializer=schema__pb2.AuthConfig.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.UpdateMTLSConfig = channel.unary_unary(
        '/immudb.schema.ImmuService/UpdateMTLSConfig',
        request_serializer=schema__pb2.MTLSConfig.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.OpenSession = channel.unary_unary(
        '/immudb.schema.ImmuService/OpenSession',
        request_serializer=schema__pb2.OpenSessionRequest.SerializeToString,
        response_deserializer=schema__pb2.OpenSessionResponse.FromString,
        )
    self.CloseSession = channel.unary_unary(
        '/immudb.schema.ImmuService/CloseSession',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.KeepAlive = channel.unary_unary(
        '/immudb.schema.ImmuService/KeepAlive',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.NewTx = channel.unary_unary(
        '/immudb.schema.ImmuService/NewTx',
        request_serializer=schema__pb2.NewTxRequest.SerializeToString,
        response_deserializer=schema__pb2.NewTxResponse.FromString,
        )
    self.Commit = channel.unary_unary(
        '/immudb.schema.ImmuService/Commit',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.CommittedSQLTx.FromString,
        )
    self.Rollback = channel.unary_unary(
        '/immudb.schema.ImmuService/Rollback',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.TxSQLExec = channel.unary_unary(
        '/immudb.schema.ImmuService/TxSQLExec',
        request_serializer=schema__pb2.SQLExecRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.TxSQLQuery = channel.unary_unary(
        '/immudb.schema.ImmuService/TxSQLQuery',
        request_serializer=schema__pb2.SQLQueryRequest.SerializeToString,
        response_deserializer=schema__pb2.SQLQueryResult.FromString,
        )
    self.Login = channel.unary_unary(
        '/immudb.schema.ImmuService/Login',
        request_serializer=schema__pb2.LoginRequest.SerializeToString,
        response_deserializer=schema__pb2.LoginResponse.FromString,
        )
    self.Logout = channel.unary_unary(
        '/immudb.schema.ImmuService/Logout',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.Set = channel.unary_unary(
        '/immudb.schema.ImmuService/Set',
        request_serializer=schema__pb2.SetRequest.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.VerifiableSet = channel.unary_unary(
        '/immudb.schema.ImmuService/VerifiableSet',
        request_serializer=schema__pb2.VerifiableSetRequest.SerializeToString,
        response_deserializer=schema__pb2.VerifiableTx.FromString,
        )
    self.Get = channel.unary_unary(
        '/immudb.schema.ImmuService/Get',
        request_serializer=schema__pb2.KeyRequest.SerializeToString,
        response_deserializer=schema__pb2.Entry.FromString,
        )
    self.VerifiableGet = channel.unary_unary(
        '/immudb.schema.ImmuService/VerifiableGet',
        request_serializer=schema__pb2.VerifiableGetRequest.SerializeToString,
        response_deserializer=schema__pb2.VerifiableEntry.FromString,
        )
    self.Delete = channel.unary_unary(
        '/immudb.schema.ImmuService/Delete',
        request_serializer=schema__pb2.DeleteKeysRequest.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.GetAll = channel.unary_unary(
        '/immudb.schema.ImmuService/GetAll',
        request_serializer=schema__pb2.KeyListRequest.SerializeToString,
        response_deserializer=schema__pb2.Entries.FromString,
        )
    self.ExecAll = channel.unary_unary(
        '/immudb.schema.ImmuService/ExecAll',
        request_serializer=schema__pb2.ExecAllRequest.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.Scan = channel.unary_unary(
        '/immudb.schema.ImmuService/Scan',
        request_serializer=schema__pb2.ScanRequest.SerializeToString,
        response_deserializer=schema__pb2.Entries.FromString,
        )
    self.Count = channel.unary_unary(
        '/immudb.schema.ImmuService/Count',
        request_serializer=schema__pb2.KeyPrefix.SerializeToString,
        response_deserializer=schema__pb2.EntryCount.FromString,
        )
    self.CountAll = channel.unary_unary(
        '/immudb.schema.ImmuService/CountAll',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.EntryCount.FromString,
        )
    self.TxById = channel.unary_unary(
        '/immudb.schema.ImmuService/TxById',
        request_serializer=schema__pb2.TxRequest.SerializeToString,
        response_deserializer=schema__pb2.Tx.FromString,
        )
    self.VerifiableTxById = channel.unary_unary(
        '/immudb.schema.ImmuService/VerifiableTxById',
        request_serializer=schema__pb2.VerifiableTxRequest.SerializeToString,
        response_deserializer=schema__pb2.VerifiableTx.FromString,
        )
    self.TxScan = channel.unary_unary(
        '/immudb.schema.ImmuService/TxScan',
        request_serializer=schema__pb2.TxScanRequest.SerializeToString,
        response_deserializer=schema__pb2.TxList.FromString,
        )
    self.History = channel.unary_unary(
        '/immudb.schema.ImmuService/History',
        request_serializer=schema__pb2.HistoryRequest.SerializeToString,
        response_deserializer=schema__pb2.Entries.FromString,
        )
    self.Health = channel.unary_unary(
        '/immudb.schema.ImmuService/Health',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.HealthResponse.FromString,
        )
    self.DatabaseHealth = channel.unary_unary(
        '/immudb.schema.ImmuService/DatabaseHealth',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.DatabaseHealthResponse.FromString,
        )
    self.CurrentState = channel.unary_unary(
        '/immudb.schema.ImmuService/CurrentState',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.ImmutableState.FromString,
        )
    self.SetReference = channel.unary_unary(
        '/immudb.schema.ImmuService/SetReference',
        request_serializer=schema__pb2.ReferenceRequest.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.VerifiableSetReference = channel.unary_unary(
        '/immudb.schema.ImmuService/VerifiableSetReference',
        request_serializer=schema__pb2.VerifiableReferenceRequest.SerializeToString,
        response_deserializer=schema__pb2.VerifiableTx.FromString,
        )
    self.ZAdd = channel.unary_unary(
        '/immudb.schema.ImmuService/ZAdd',
        request_serializer=schema__pb2.ZAddRequest.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.VerifiableZAdd = channel.unary_unary(
        '/immudb.schema.ImmuService/VerifiableZAdd',
        request_serializer=schema__pb2.VerifiableZAddRequest.SerializeToString,
        response_deserializer=schema__pb2.VerifiableTx.FromString,
        )
    self.ZScan = channel.unary_unary(
        '/immudb.schema.ImmuService/ZScan',
        request_serializer=schema__pb2.ZScanRequest.SerializeToString,
        response_deserializer=schema__pb2.ZEntries.FromString,
        )
    self.CreateDatabase = channel.unary_unary(
        '/immudb.schema.ImmuService/CreateDatabase',
        request_serializer=schema__pb2.Database.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.CreateDatabaseWith = channel.unary_unary(
        '/immudb.schema.ImmuService/CreateDatabaseWith',
        request_serializer=schema__pb2.DatabaseSettings.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.CreateDatabaseV2 = channel.unary_unary(
        '/immudb.schema.ImmuService/CreateDatabaseV2',
        request_serializer=schema__pb2.CreateDatabaseRequest.SerializeToString,
        response_deserializer=schema__pb2.CreateDatabaseResponse.FromString,
        )
    self.LoadDatabase = channel.unary_unary(
        '/immudb.schema.ImmuService/LoadDatabase',
        request_serializer=schema__pb2.LoadDatabaseRequest.SerializeToString,
        response_deserializer=schema__pb2.LoadDatabaseResponse.FromString,
        )
    self.UnloadDatabase = channel.unary_unary(
        '/immudb.schema.ImmuService/UnloadDatabase',
        request_serializer=schema__pb2.UnloadDatabaseRequest.SerializeToString,
        response_deserializer=schema__pb2.UnloadDatabaseResponse.FromString,
        )
    self.DeleteDatabase = channel.unary_unary(
        '/immudb.schema.ImmuService/DeleteDatabase',
        request_serializer=schema__pb2.DeleteDatabaseRequest.SerializeToString,
        response_deserializer=schema__pb2.DeleteDatabaseResponse.FromString,
        )
    self.DatabaseList = channel.unary_unary(
        '/immudb.schema.ImmuService/DatabaseList',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.DatabaseListResponse.FromString,
        )
    self.DatabaseListV2 = channel.unary_unary(
        '/immudb.schema.ImmuService/DatabaseListV2',
        request_serializer=schema__pb2.DatabaseListRequestV2.SerializeToString,
        response_deserializer=schema__pb2.DatabaseListResponseV2.FromString,
        )
    self.UseDatabase = channel.unary_unary(
        '/immudb.schema.ImmuService/UseDatabase',
        request_serializer=schema__pb2.Database.SerializeToString,
        response_deserializer=schema__pb2.UseDatabaseReply.FromString,
        )
    self.UpdateDatabase = channel.unary_unary(
        '/immudb.schema.ImmuService/UpdateDatabase',
        request_serializer=schema__pb2.DatabaseSettings.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.UpdateDatabaseV2 = channel.unary_unary(
        '/immudb.schema.ImmuService/UpdateDatabaseV2',
        request_serializer=schema__pb2.UpdateDatabaseRequest.SerializeToString,
        response_deserializer=schema__pb2.UpdateDatabaseResponse.FromString,
        )
    self.GetDatabaseSettings = channel.unary_unary(
        '/immudb.schema.ImmuService/GetDatabaseSettings',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.DatabaseSettings.FromString,
        )
    self.GetDatabaseSettingsV2 = channel.unary_unary(
        '/immudb.schema.ImmuService/GetDatabaseSettingsV2',
        request_serializer=schema__pb2.DatabaseSettingsRequest.SerializeToString,
        response_deserializer=schema__pb2.DatabaseSettingsResponse.FromString,
        )
    self.FlushIndex = channel.unary_unary(
        '/immudb.schema.ImmuService/FlushIndex',
        request_serializer=schema__pb2.FlushIndexRequest.SerializeToString,
        response_deserializer=schema__pb2.FlushIndexResponse.FromString,
        )
    self.CompactIndex = channel.unary_unary(
        '/immudb.schema.ImmuService/CompactIndex',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.streamGet = channel.unary_stream(
        '/immudb.schema.ImmuService/streamGet',
        request_serializer=schema__pb2.KeyRequest.SerializeToString,
        response_deserializer=schema__pb2.Chunk.FromString,
        )
    self.streamSet = channel.stream_unary(
        '/immudb.schema.ImmuService/streamSet',
        request_serializer=schema__pb2.Chunk.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.streamVerifiableGet = channel.unary_stream(
        '/immudb.schema.ImmuService/streamVerifiableGet',
        request_serializer=schema__pb2.VerifiableGetRequest.SerializeToString,
        response_deserializer=schema__pb2.Chunk.FromString,
        )
    self.streamVerifiableSet = channel.stream_unary(
        '/immudb.schema.ImmuService/streamVerifiableSet',
        request_serializer=schema__pb2.Chunk.SerializeToString,
        response_deserializer=schema__pb2.VerifiableTx.FromString,
        )
    self.streamScan = channel.unary_stream(
        '/immudb.schema.ImmuService/streamScan',
        request_serializer=schema__pb2.ScanRequest.SerializeToString,
        response_deserializer=schema__pb2.Chunk.FromString,
        )
    self.streamZScan = channel.unary_stream(
        '/immudb.schema.ImmuService/streamZScan',
        request_serializer=schema__pb2.ZScanRequest.SerializeToString,
        response_deserializer=schema__pb2.Chunk.FromString,
        )
    self.streamHistory = channel.unary_stream(
        '/immudb.schema.ImmuService/streamHistory',
        request_serializer=schema__pb2.HistoryRequest.SerializeToString,
        response_deserializer=schema__pb2.Chunk.FromString,
        )
    self.streamExecAll = channel.stream_unary(
        '/immudb.schema.ImmuService/streamExecAll',
        request_serializer=schema__pb2.Chunk.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.exportTx = channel.unary_stream(
        '/immudb.schema.ImmuService/exportTx',
        request_serializer=schema__pb2.ExportTxRequest.SerializeToString,
        response_deserializer=schema__pb2.Chunk.FromString,
        )
    self.replicateTx = channel.stream_unary(
        '/immudb.schema.ImmuService/replicateTx',
        request_serializer=schema__pb2.Chunk.SerializeToString,
        response_deserializer=schema__pb2.TxHeader.FromString,
        )
    self.SQLExec = channel.unary_unary(
        '/immudb.schema.ImmuService/SQLExec',
        request_serializer=schema__pb2.SQLExecRequest.SerializeToString,
        response_deserializer=schema__pb2.SQLExecResult.FromString,
        )
    self.SQLQuery = channel.unary_unary(
        '/immudb.schema.ImmuService/SQLQuery',
        request_serializer=schema__pb2.SQLQueryRequest.SerializeToString,
        response_deserializer=schema__pb2.SQLQueryResult.FromString,
        )
    self.ListTables = channel.unary_unary(
        '/immudb.schema.ImmuService/ListTables',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=schema__pb2.SQLQueryResult.FromString,
        )
    self.DescribeTable = channel.unary_unary(
        '/immudb.schema.ImmuService/DescribeTable',
        request_serializer=schema__pb2.Table.SerializeToString,
        response_deserializer=schema__pb2.SQLQueryResult.FromString,
        )
    self.VerifiableSQLGet = channel.unary_unary(
        '/immudb.schema.ImmuService/VerifiableSQLGet',
        request_serializer=schema__pb2.VerifiableSQLGetRequest.SerializeToString,
        response_deserializer=schema__pb2.VerifiableSQLEntry.FromString,
        )


class ImmuServiceServicer(object):
  """immudb gRPC & REST service
  """

  def ListUsers(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateUser(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ChangePassword(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ChangePermission(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetActiveUser(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateAuthConfig(self, request, context):
    """DEPRECATED
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateMTLSConfig(self, request, context):
    """DEPRECATED
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def OpenSession(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CloseSession(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def KeepAlive(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def NewTx(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Commit(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Rollback(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def TxSQLExec(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def TxSQLQuery(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Login(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Logout(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Set(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def VerifiableSet(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Get(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def VerifiableGet(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Delete(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAll(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ExecAll(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Scan(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Count(self, request, context):
    """NOT YET SUPPORTED
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CountAll(self, request, context):
    """NOT YET SUPPORTED
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def TxById(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def VerifiableTxById(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def TxScan(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def History(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Health(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DatabaseHealth(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CurrentState(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetReference(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def VerifiableSetReference(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ZAdd(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def VerifiableZAdd(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ZScan(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateDatabase(self, request, context):
    """DEPRECATED: kept for backward compatibility
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateDatabaseWith(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateDatabaseV2(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def LoadDatabase(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UnloadDatabase(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteDatabase(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DatabaseList(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DatabaseListV2(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UseDatabase(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateDatabase(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateDatabaseV2(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetDatabaseSettings(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetDatabaseSettingsV2(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FlushIndex(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CompactIndex(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamGet(self, request, context):
    """Streams
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamSet(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamVerifiableGet(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamVerifiableSet(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamScan(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamZScan(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamHistory(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def streamExecAll(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def exportTx(self, request, context):
    """Replication
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def replicateTx(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SQLExec(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SQLQuery(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListTables(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DescribeTable(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def VerifiableSQLGet(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ImmuServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ListUsers': grpc.unary_unary_rpc_method_handler(
          servicer.ListUsers,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.UserList.SerializeToString,
      ),
      'CreateUser': grpc.unary_unary_rpc_method_handler(
          servicer.CreateUser,
          request_deserializer=schema__pb2.CreateUserRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'ChangePassword': grpc.unary_unary_rpc_method_handler(
          servicer.ChangePassword,
          request_deserializer=schema__pb2.ChangePasswordRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'ChangePermission': grpc.unary_unary_rpc_method_handler(
          servicer.ChangePermission,
          request_deserializer=schema__pb2.ChangePermissionRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'SetActiveUser': grpc.unary_unary_rpc_method_handler(
          servicer.SetActiveUser,
          request_deserializer=schema__pb2.SetActiveUserRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'UpdateAuthConfig': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateAuthConfig,
          request_deserializer=schema__pb2.AuthConfig.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'UpdateMTLSConfig': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateMTLSConfig,
          request_deserializer=schema__pb2.MTLSConfig.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'OpenSession': grpc.unary_unary_rpc_method_handler(
          servicer.OpenSession,
          request_deserializer=schema__pb2.OpenSessionRequest.FromString,
          response_serializer=schema__pb2.OpenSessionResponse.SerializeToString,
      ),
      'CloseSession': grpc.unary_unary_rpc_method_handler(
          servicer.CloseSession,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'KeepAlive': grpc.unary_unary_rpc_method_handler(
          servicer.KeepAlive,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'NewTx': grpc.unary_unary_rpc_method_handler(
          servicer.NewTx,
          request_deserializer=schema__pb2.NewTxRequest.FromString,
          response_serializer=schema__pb2.NewTxResponse.SerializeToString,
      ),
      'Commit': grpc.unary_unary_rpc_method_handler(
          servicer.Commit,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.CommittedSQLTx.SerializeToString,
      ),
      'Rollback': grpc.unary_unary_rpc_method_handler(
          servicer.Rollback,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'TxSQLExec': grpc.unary_unary_rpc_method_handler(
          servicer.TxSQLExec,
          request_deserializer=schema__pb2.SQLExecRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'TxSQLQuery': grpc.unary_unary_rpc_method_handler(
          servicer.TxSQLQuery,
          request_deserializer=schema__pb2.SQLQueryRequest.FromString,
          response_serializer=schema__pb2.SQLQueryResult.SerializeToString,
      ),
      'Login': grpc.unary_unary_rpc_method_handler(
          servicer.Login,
          request_deserializer=schema__pb2.LoginRequest.FromString,
          response_serializer=schema__pb2.LoginResponse.SerializeToString,
      ),
      'Logout': grpc.unary_unary_rpc_method_handler(
          servicer.Logout,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'Set': grpc.unary_unary_rpc_method_handler(
          servicer.Set,
          request_deserializer=schema__pb2.SetRequest.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'VerifiableSet': grpc.unary_unary_rpc_method_handler(
          servicer.VerifiableSet,
          request_deserializer=schema__pb2.VerifiableSetRequest.FromString,
          response_serializer=schema__pb2.VerifiableTx.SerializeToString,
      ),
      'Get': grpc.unary_unary_rpc_method_handler(
          servicer.Get,
          request_deserializer=schema__pb2.KeyRequest.FromString,
          response_serializer=schema__pb2.Entry.SerializeToString,
      ),
      'VerifiableGet': grpc.unary_unary_rpc_method_handler(
          servicer.VerifiableGet,
          request_deserializer=schema__pb2.VerifiableGetRequest.FromString,
          response_serializer=schema__pb2.VerifiableEntry.SerializeToString,
      ),
      'Delete': grpc.unary_unary_rpc_method_handler(
          servicer.Delete,
          request_deserializer=schema__pb2.DeleteKeysRequest.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'GetAll': grpc.unary_unary_rpc_method_handler(
          servicer.GetAll,
          request_deserializer=schema__pb2.KeyListRequest.FromString,
          response_serializer=schema__pb2.Entries.SerializeToString,
      ),
      'ExecAll': grpc.unary_unary_rpc_method_handler(
          servicer.ExecAll,
          request_deserializer=schema__pb2.ExecAllRequest.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'Scan': grpc.unary_unary_rpc_method_handler(
          servicer.Scan,
          request_deserializer=schema__pb2.ScanRequest.FromString,
          response_serializer=schema__pb2.Entries.SerializeToString,
      ),
      'Count': grpc.unary_unary_rpc_method_handler(
          servicer.Count,
          request_deserializer=schema__pb2.KeyPrefix.FromString,
          response_serializer=schema__pb2.EntryCount.SerializeToString,
      ),
      'CountAll': grpc.unary_unary_rpc_method_handler(
          servicer.CountAll,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.EntryCount.SerializeToString,
      ),
      'TxById': grpc.unary_unary_rpc_method_handler(
          servicer.TxById,
          request_deserializer=schema__pb2.TxRequest.FromString,
          response_serializer=schema__pb2.Tx.SerializeToString,
      ),
      'VerifiableTxById': grpc.unary_unary_rpc_method_handler(
          servicer.VerifiableTxById,
          request_deserializer=schema__pb2.VerifiableTxRequest.FromString,
          response_serializer=schema__pb2.VerifiableTx.SerializeToString,
      ),
      'TxScan': grpc.unary_unary_rpc_method_handler(
          servicer.TxScan,
          request_deserializer=schema__pb2.TxScanRequest.FromString,
          response_serializer=schema__pb2.TxList.SerializeToString,
      ),
      'History': grpc.unary_unary_rpc_method_handler(
          servicer.History,
          request_deserializer=schema__pb2.HistoryRequest.FromString,
          response_serializer=schema__pb2.Entries.SerializeToString,
      ),
      'Health': grpc.unary_unary_rpc_method_handler(
          servicer.Health,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.HealthResponse.SerializeToString,
      ),
      'DatabaseHealth': grpc.unary_unary_rpc_method_handler(
          servicer.DatabaseHealth,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.DatabaseHealthResponse.SerializeToString,
      ),
      'CurrentState': grpc.unary_unary_rpc_method_handler(
          servicer.CurrentState,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.ImmutableState.SerializeToString,
      ),
      'SetReference': grpc.unary_unary_rpc_method_handler(
          servicer.SetReference,
          request_deserializer=schema__pb2.ReferenceRequest.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'VerifiableSetReference': grpc.unary_unary_rpc_method_handler(
          servicer.VerifiableSetReference,
          request_deserializer=schema__pb2.VerifiableReferenceRequest.FromString,
          response_serializer=schema__pb2.VerifiableTx.SerializeToString,
      ),
      'ZAdd': grpc.unary_unary_rpc_method_handler(
          servicer.ZAdd,
          request_deserializer=schema__pb2.ZAddRequest.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'VerifiableZAdd': grpc.unary_unary_rpc_method_handler(
          servicer.VerifiableZAdd,
          request_deserializer=schema__pb2.VerifiableZAddRequest.FromString,
          response_serializer=schema__pb2.VerifiableTx.SerializeToString,
      ),
      'ZScan': grpc.unary_unary_rpc_method_handler(
          servicer.ZScan,
          request_deserializer=schema__pb2.ZScanRequest.FromString,
          response_serializer=schema__pb2.ZEntries.SerializeToString,
      ),
      'CreateDatabase': grpc.unary_unary_rpc_method_handler(
          servicer.CreateDatabase,
          request_deserializer=schema__pb2.Database.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'CreateDatabaseWith': grpc.unary_unary_rpc_method_handler(
          servicer.CreateDatabaseWith,
          request_deserializer=schema__pb2.DatabaseSettings.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'CreateDatabaseV2': grpc.unary_unary_rpc_method_handler(
          servicer.CreateDatabaseV2,
          request_deserializer=schema__pb2.CreateDatabaseRequest.FromString,
          response_serializer=schema__pb2.CreateDatabaseResponse.SerializeToString,
      ),
      'LoadDatabase': grpc.unary_unary_rpc_method_handler(
          servicer.LoadDatabase,
          request_deserializer=schema__pb2.LoadDatabaseRequest.FromString,
          response_serializer=schema__pb2.LoadDatabaseResponse.SerializeToString,
      ),
      'UnloadDatabase': grpc.unary_unary_rpc_method_handler(
          servicer.UnloadDatabase,
          request_deserializer=schema__pb2.UnloadDatabaseRequest.FromString,
          response_serializer=schema__pb2.UnloadDatabaseResponse.SerializeToString,
      ),
      'DeleteDatabase': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteDatabase,
          request_deserializer=schema__pb2.DeleteDatabaseRequest.FromString,
          response_serializer=schema__pb2.DeleteDatabaseResponse.SerializeToString,
      ),
      'DatabaseList': grpc.unary_unary_rpc_method_handler(
          servicer.DatabaseList,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.DatabaseListResponse.SerializeToString,
      ),
      'DatabaseListV2': grpc.unary_unary_rpc_method_handler(
          servicer.DatabaseListV2,
          request_deserializer=schema__pb2.DatabaseListRequestV2.FromString,
          response_serializer=schema__pb2.DatabaseListResponseV2.SerializeToString,
      ),
      'UseDatabase': grpc.unary_unary_rpc_method_handler(
          servicer.UseDatabase,
          request_deserializer=schema__pb2.Database.FromString,
          response_serializer=schema__pb2.UseDatabaseReply.SerializeToString,
      ),
      'UpdateDatabase': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateDatabase,
          request_deserializer=schema__pb2.DatabaseSettings.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'UpdateDatabaseV2': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateDatabaseV2,
          request_deserializer=schema__pb2.UpdateDatabaseRequest.FromString,
          response_serializer=schema__pb2.UpdateDatabaseResponse.SerializeToString,
      ),
      'GetDatabaseSettings': grpc.unary_unary_rpc_method_handler(
          servicer.GetDatabaseSettings,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.DatabaseSettings.SerializeToString,
      ),
      'GetDatabaseSettingsV2': grpc.unary_unary_rpc_method_handler(
          servicer.GetDatabaseSettingsV2,
          request_deserializer=schema__pb2.DatabaseSettingsRequest.FromString,
          response_serializer=schema__pb2.DatabaseSettingsResponse.SerializeToString,
      ),
      'FlushIndex': grpc.unary_unary_rpc_method_handler(
          servicer.FlushIndex,
          request_deserializer=schema__pb2.FlushIndexRequest.FromString,
          response_serializer=schema__pb2.FlushIndexResponse.SerializeToString,
      ),
      'CompactIndex': grpc.unary_unary_rpc_method_handler(
          servicer.CompactIndex,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'streamGet': grpc.unary_stream_rpc_method_handler(
          servicer.streamGet,
          request_deserializer=schema__pb2.KeyRequest.FromString,
          response_serializer=schema__pb2.Chunk.SerializeToString,
      ),
      'streamSet': grpc.stream_unary_rpc_method_handler(
          servicer.streamSet,
          request_deserializer=schema__pb2.Chunk.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'streamVerifiableGet': grpc.unary_stream_rpc_method_handler(
          servicer.streamVerifiableGet,
          request_deserializer=schema__pb2.VerifiableGetRequest.FromString,
          response_serializer=schema__pb2.Chunk.SerializeToString,
      ),
      'streamVerifiableSet': grpc.stream_unary_rpc_method_handler(
          servicer.streamVerifiableSet,
          request_deserializer=schema__pb2.Chunk.FromString,
          response_serializer=schema__pb2.VerifiableTx.SerializeToString,
      ),
      'streamScan': grpc.unary_stream_rpc_method_handler(
          servicer.streamScan,
          request_deserializer=schema__pb2.ScanRequest.FromString,
          response_serializer=schema__pb2.Chunk.SerializeToString,
      ),
      'streamZScan': grpc.unary_stream_rpc_method_handler(
          servicer.streamZScan,
          request_deserializer=schema__pb2.ZScanRequest.FromString,
          response_serializer=schema__pb2.Chunk.SerializeToString,
      ),
      'streamHistory': grpc.unary_stream_rpc_method_handler(
          servicer.streamHistory,
          request_deserializer=schema__pb2.HistoryRequest.FromString,
          response_serializer=schema__pb2.Chunk.SerializeToString,
      ),
      'streamExecAll': grpc.stream_unary_rpc_method_handler(
          servicer.streamExecAll,
          request_deserializer=schema__pb2.Chunk.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'exportTx': grpc.unary_stream_rpc_method_handler(
          servicer.exportTx,
          request_deserializer=schema__pb2.ExportTxRequest.FromString,
          response_serializer=schema__pb2.Chunk.SerializeToString,
      ),
      'replicateTx': grpc.stream_unary_rpc_method_handler(
          servicer.replicateTx,
          request_deserializer=schema__pb2.Chunk.FromString,
          response_serializer=schema__pb2.TxHeader.SerializeToString,
      ),
      'SQLExec': grpc.unary_unary_rpc_method_handler(
          servicer.SQLExec,
          request_deserializer=schema__pb2.SQLExecRequest.FromString,
          response_serializer=schema__pb2.SQLExecResult.SerializeToString,
      ),
      'SQLQuery': grpc.unary_unary_rpc_method_handler(
          servicer.SQLQuery,
          request_deserializer=schema__pb2.SQLQueryRequest.FromString,
          response_serializer=schema__pb2.SQLQueryResult.SerializeToString,
      ),
      'ListTables': grpc.unary_unary_rpc_method_handler(
          servicer.ListTables,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=schema__pb2.SQLQueryResult.SerializeToString,
      ),
      'DescribeTable': grpc.unary_unary_rpc_method_handler(
          servicer.DescribeTable,
          request_deserializer=schema__pb2.Table.FromString,
          response_serializer=schema__pb2.SQLQueryResult.SerializeToString,
      ),
      'VerifiableSQLGet': grpc.unary_unary_rpc_method_handler(
          servicer.VerifiableSQLGet,
          request_deserializer=schema__pb2.VerifiableSQLGetRequest.FromString,
          response_serializer=schema__pb2.VerifiableSQLEntry.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'immudb.schema.ImmuService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from immudb.grpc import schema_pb2_grpc
from immudb import datatypes
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from immudb import constants
from immudb import grpcutils
from immudb.handler.sqlquery import _call_with_executor as executeSQLQuery
from immudb.handler.sqlexec import _call_with_executor as executeSQLExec


class Tx:
    def __init__(self, stub, session, channel):
        self.stub = stub
        self.session = session
        self.channel = channel
        self.txStub = None

    def makeTransactionInterceptedStub(self, transactionResponse):
        transactionId = transactionResponse.transactionID
        sessionId = self.session.sessionID
        headersInterceptors = [grpcutils.header_adder_interceptor(
            'sessionid', sessionId), grpcutils.header_adder_interceptor('transactionid', transactionId)]
        interceptedChannel, stub = grpcutils.get_intercepted_stub(
            self.channel, headersInterceptors)
        return stub

    def newTx(self, mode=datatypes.TxMode.ReadWrite):
        req = schema_pb2_grpc.schema__pb2.NewTxRequest(mode=mode)
        resp = self.stub.NewTx(req)
        self.txStub = self.makeTransactionInterceptedStub(resp)
        return self

    def commit(self):
        resp = self.txStub.Commit(google_dot_protobuf_dot_empty__pb2.Empty())
        self.txStub = None
        return resp

    def rollback(self):
        resp = self.txStub.Rollback(google_dot_protobuf_dot_empty__pb2.Empty())
        self.txStub = None
        return resp

    def sqlQuery(self, query, params=dict(), columnNameMode=constants.COLUMN_NAME_MODE_NONE):
        return executeSQLQuery(query, params, columnNameMode, self.txStub.TxSQLQuery)

    def sqlExec(self, stmt, params=dict(), noWait=False):
        return executeSQLExec(stmt, params, noWait, self.txStub.TxSQLExec)



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb import datatypes


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key: bytes, offset: int, limit: int, desc: bool):
    state = rs.get()
    request = schema_pb2_grpc.schema__pb2.HistoryRequest(
        key=key,
        offset=offset,
        limit=limit,
        desc=desc,
        sinceTx=state.txId
    )
    histo = service.History(request)
    histolist = []
    for i in histo.entries:
        histolist.append(datatypes.historyResponseItem(
            key=i.key,
            value=i.value,
            tx=i.tx
        ))
    return histolist



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interceptor that adds headers to outgoing requests."""

import collections

import grpc
import immudb.generic_client_interceptor as generic_client_interceptor
import immudb.grpc.schema_pb2_grpc as schema_pb2_grpc


class _ClientCallDetails(
        collections.namedtuple(
            '_ClientCallDetails',
            ('method', 'timeout', 'metadata', 'credentials')),
        grpc.ClientCallDetails):
    pass


def timeout_adder_interceptor(stubTimeout=None):

    def intercept_call(client_call_details, request_iterator, request_streaming,
                       response_streaming):
        timeoutToSet = client_call_details.timeout
        if(timeoutToSet == None):
            timeoutToSet = stubTimeout
        client_call_details = _ClientCallDetails(
            client_call_details.method, timeoutToSet, client_call_details.metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return generic_client_interceptor.create(intercept_call)


def header_adder_interceptor(header, value):

    def intercept_call(client_call_details, request_iterator, request_streaming,
                       response_streaming):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.append((
            header,
            value,
        ))
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return generic_client_interceptor.create(intercept_call)


def get_intercepted_stub(channel, headers):
    try:
        intercepted = grpc.intercept_channel(channel, *headers)
        newStub = schema_pb2_grpc.ImmuServiceStub(intercepted)
        return intercepted, newStub
    except ValueError as e:
        raise Exception(
            "Attempted to login on termninated client, channel has been shutdown") from e



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2

from immudb import grpcutils
from immudb.handler import (batchGet, batchSet, changePassword, changePermission, createUser,
                            currentRoot, createDatabase, databaseList, deleteKeys, useDatabase,
                            get, listUsers, sqldescribe, verifiedGet, verifiedSet, setValue, history,
                            scan, reference, verifiedreference, zadd, verifiedzadd,
                            zscan, healthcheck, health, txbyid, verifiedtxbyid, sqlexec, sqlquery,
                            listtables, execAll, transaction)
from immudb.rootService import *
from immudb.grpc import schema_pb2_grpc
import warnings
import ecdsa
from immudb.datatypes import DeleteKeysRequest
from immudb.embedded.store import KVMetadata
import threading
import queue

import datetime


class ImmudbClient:

    def __init__(self, immudUrl=None, rs: RootService = None, publicKeyFile: str = None, timeout=None):
        if immudUrl is None:
            immudUrl = "localhost:3322"
        self.timeout = timeout
        self.channel = grpc.insecure_channel(immudUrl)
        self._resetStub()
        if rs is None:
            self.__rs = RootService()
        else:
            self.__rs = rs
        self.__url = immudUrl
        self.loadKey(publicKeyFile)
        self.__login_response = None
        self._session_response = None

    def loadKey(self, kfile: str):
        if kfile is None:
            self.__vk = None
        else:
            with open(kfile) as f:
                self.__vk = ecdsa.VerifyingKey.from_pem(f.read())

    def shutdown(self):
        self.channel.close()
        self.channel = None
        self.intercept_channel.close
        self.intercept_channel = None
        self.__rs = None

    def set_session_id_interceptor(self, openSessionResponse):
        sessionId = openSessionResponse.sessionID
        self.headersInterceptors = [
            grpcutils.header_adder_interceptor('sessionid', sessionId)]
        return self.get_intercepted_stub()

    def set_token_header_interceptor(self, response):
        try:
            token = response.token
        except AttributeError:
            token = response.reply.token
        self.headersInterceptors = [
            grpcutils.header_adder_interceptor(
                'authorization', "Bearer " + token
            )
        ]
        return self.get_intercepted_stub()

    def get_intercepted_stub(self):
        allInterceptors = self.headersInterceptors + self.clientInterceptors
        intercepted, newStub = grpcutils.get_intercepted_stub(
            self.channel, allInterceptors)
        self.intercept_channel = intercepted
        return newStub

    @property
    def stub(self):
        return self.__stub

# from here on same order as in Golang ImmuClient interface (pkg/client/client.go)

    # Not implemented: disconnect
    # Not implemented: isConnected
    # Not implemented: waitForHealthCheck
    def healthCheck(self):
        return healthcheck.call(self.__stub, self.__rs)

    # Not implemented: connect
    def _convertToBytes(self, what):
        if(type(what) != bytes):
            return bytes(what, encoding='utf-8')
        return what

    def login(self, username, password, database=b"defaultdb"):
        convertedUsername = self._convertToBytes(username)
        convertedPassword = self._convertToBytes(password)
        convertedDatabase = self._convertToBytes(database)
        req = schema_pb2_grpc.schema__pb2.LoginRequest(
            user=convertedUsername, password=convertedPassword)
        try:
            self.__login_response = schema_pb2_grpc.schema__pb2.LoginResponse = \
                self.__stub.Login(
                    req
                )
        except ValueError as e:
            raise Exception(
                "Attempted to login on termninated client, channel has been shutdown") from e

        self.__stub = self.set_token_header_interceptor(self.__login_response)
        # Select database, modifying stub function accordingly
        request = schema_pb2_grpc.schema__pb2.Database(
            databaseName=convertedDatabase)
        resp = self.__stub.UseDatabase(request)
        self.__stub = self.set_token_header_interceptor(resp)

        self.__rs.init("{}/{}".format(self.__url, database), self.__stub)
        return self.__login_response

    def logout(self):
        self.__stub.Logout(google_dot_protobuf_dot_empty__pb2.Empty())
        self.__login_response = None
        self._resetStub()

    def _resetStub(self):
        self.headersInterceptors = []
        self.clientInterceptors = []
        if(self.timeout != None):
            self.clientInterceptors.append(
                grpcutils.timeout_adder_interceptor(self.timeout))
        self.__stub = schema_pb2_grpc.ImmuServiceStub(self.channel)
        self.__stub = self.get_intercepted_stub()

    def keepAlive(self):
        self.__stub.KeepAlive(google_dot_protobuf_dot_empty__pb2.Empty())

    def openManagedSession(self, username, password, database=b"defaultdb", keepAliveInterval=60):
        class ManagedSession:
            def __init__(this, keepAliveInterval):
                this.keepAliveInterval = keepAliveInterval
                this.keepAliveStarted = False
                this.keepAliveProcess = None
                this.queue = queue.Queue()

            def manage(this):
                while this.keepAliveStarted:
                    try:
                        what = this.queue.get(True, this.keepAliveInterval)
                    except queue.Empty:
                        self.keepAlive()

            def __enter__(this):
                interface = self.openSession(username, password, database)
                this.keepAliveStarted = True
                this.keepAliveProcess = threading.Thread(target=this.manage)
                this.keepAliveProcess.start()
                return interface

            def __exit__(this, type, value, traceback):
                this.keepAliveStarted = False
                this.queue.put(b'0')
                self.closeSession()

        return ManagedSession(keepAliveInterval)

    def openSession(self, username, password, database=b"defaultdb"):
        convertedUsername = self._convertToBytes(username)
        convertedPassword = self._convertToBytes(password)
        convertedDatabase = self._convertToBytes(database)
        req = schema_pb2_grpc.schema__pb2.OpenSessionRequest(
            username=convertedUsername,
            password=convertedPassword,
            databaseName=convertedDatabase
        )
        self._session_response = self.__stub.OpenSession(
            req)
        self.__stub = self.set_session_id_interceptor(self._session_response)
        return transaction.Tx(self.__stub, self._session_response, self.channel)

    def closeSession(self):
        self.__stub.CloseSession(google_dot_protobuf_dot_empty__pb2.Empty())
        self._session_response = None
        self._resetStub()

    def createUser(self, user, password, permission, database):
        request = schema_pb2_grpc.schema__pb2.CreateUserRequest(
            user=bytes(user, encoding='utf-8'),
            password=bytes(password, encoding='utf-8'),
            permission=permission,
            database=database
        )
        return createUser.call(self.__stub, self.__rs, request)

    def listUsers(self):
        return listUsers.call(self.__stub, None)

    def changePassword(self, user, newPassword, oldPassword):
        request = schema_pb2_grpc.schema__pb2.ChangePasswordRequest(
            user=bytes(user, encoding='utf-8'),
            newPassword=bytes(newPassword, encoding='utf-8'),
            oldPassword=bytes(oldPassword, encoding='utf-8')
        )
        return changePassword.call(self.__stub, self.__rs, request)

    def changePermission(self, action, user, database, permissions):
        return changePermission.call(self.__stub, self.__rs, action, user, database, permissions)

    # Not implemented: updateAuthConfig
    # Not implemented: updateMTLSConfig

    # Not implemented: with[.*]

    # Not implemented: getServiceClient
    # Not implemented: getOptions
    # Not implemented: setupDialOptions

    def databaseList(self):
        dbs = databaseList.call(self.__stub, self.__rs, None)
        return [x.databaseName for x in dbs.dblist.databases]

    # Not implemented: databaseListV2

    def createDatabase(self, dbName: bytes):
        request = schema_pb2_grpc.schema__pb2.Database(databaseName=dbName)
        return createDatabase.call(self.__stub, self.__rs, request)

    # Not implemented: createDatabaseV2
    # Not implemented: loadDatabase
    # Not implemented: unloadDatabase
    # Not implemented: deleteDatabase

    def useDatabase(self, dbName: bytes):
        request = schema_pb2_grpc.schema__pb2.Database(databaseName=dbName)
        resp = useDatabase.call(self.__stub, self.__rs, request)
        # modify header token accordingly
        self.__stub = self.set_token_header_interceptor(resp)
        self.__rs.init(dbName, self.__stub)
        return resp

    # Not implemented: updateDatabase
    # Not implemented: updateDatabaseV2
    # Not implemented: getDatabaseSettings
    # Not implemented: getDatabaseSettingsV2

    # Not implemented: setActiveUser

    # Not implemented: flushIndex

    def compactIndex(self):
        self.__stub.CompactIndex(google_dot_protobuf_dot_empty__pb2.Empty())

    def health(self):
        return health.call(self.__stub, self.__rs)

    def currentState(self):
        return currentRoot.call(self.__stub, self.__rs, None)

    def set(self, key: bytes, value: bytes):
        return setValue.call(self.__stub, self.__rs, key, value)

    def verifiedSet(self, key: bytes, value: bytes):
        return verifiedSet.call(self.__stub, self.__rs, key, value, self.__vk)

    def expireableSet(self, key: bytes, value: bytes, expiresAt: datetime.datetime):
        metadata = KVMetadata()
        metadata.ExpiresAt(expiresAt)
        return setValue.call(self.__stub, self.__rs, key, value, metadata)

    def get(self, key: bytes, atRevision: int = None):
        return get.call(self.__stub, self.__rs, key, atRevision=atRevision)

    # Not implemented: getSince
    # Not implemented: getAt

    def verifiedGet(self, key: bytes, atRevision: int = None):
        return verifiedGet.call(self.__stub, self.__rs, key, verifying_key=self.__vk, atRevision=atRevision)

    def verifiedGetSince(self, key: bytes, sinceTx: int):
        return verifiedGet.call(self.__stub, self.__rs, key, sinceTx=sinceTx, verifying_key=self.__vk)

    def verifiedGetAt(self, key: bytes, atTx: int):
        return verifiedGet.call(self.__stub, self.__rs, key, atTx, self.__vk)

    def history(self, key: bytes, offset: int, limit: int, sortorder: bool):
        return history.call(self.__stub, self.__rs, key, offset, limit, sortorder)

    def zAdd(self, zset: bytes, score: float, key: bytes, atTx: int = 0):
        return zadd.call(self.__stub, self.__rs, zset, score, key, atTx)

    def verifiedZAdd(self, zset: bytes, score: float, key: bytes, atTx: int = 0):
        return verifiedzadd.call(self.__stub, self.__rs, zset, score, key, atTx, self.__vk)

    # Not implemented: zAddAt
    # Not implemented: verifiedZAddAt

    def scan(self, key: bytes, prefix: bytes, desc: bool, limit: int, sinceTx: int = None):
        return scan.call(self.__stub, self.__rs, key, prefix, desc, limit, sinceTx)

    def zScan(self, zset: bytes, seekKey: bytes, seekScore: float,
              seekAtTx: int, inclusive: bool, limit: int, desc: bool, minscore: float,
              maxscore: float, sinceTx=None, nowait=False):
        return zscan.call(self.__stub, self.__rs, zset, seekKey, seekScore,
                          seekAtTx, inclusive, limit, desc, minscore,
                          maxscore, sinceTx, nowait)

    def txById(self, tx: int):
        return txbyid.call(self.__stub, self.__rs, tx)

    def verifiedTxById(self, tx: int):
        return verifiedtxbyid.call(self.__stub, self.__rs, tx, self.__vk)

    # Not implemented: txByIDWithSpec

    # Not implemented: txScan

    # Not implemented: count
    # Not implemented: countAll

    def setAll(self, kv: dict):
        return batchSet.call(self.__stub, self.__rs, kv)

    def getAll(self, keys: list):
        resp = batchGet.call(self.__stub, self.__rs, keys)
        return {key: value.value for key, value in resp.items()}

    def delete(self, req: DeleteKeysRequest):
        return deleteKeys.call(self.__stub, req)

    def execAll(self, ops: list, noWait=False):
        return execAll.call(self.__stub, self.__rs, ops, noWait)

    def setReference(self, referredkey: bytes, newkey:  bytes):
        return reference.call(self.__stub, self.__rs, referredkey, newkey)

    def verifiedSetReference(self, referredkey: bytes, newkey:  bytes):
        return verifiedreference.call(self.__stub, self.__rs, referredkey, newkey, verifying_key=self.__vk)

    # Not implemented: setReferenceAt
    # Not implemented: verifiedSetReferenceAt

    # Not implemented: dump

    # Not implemented: stream[.*]

    # Not implemented: exportTx
    # Not implemented: replicateTx

    def sqlExec(self, stmt, params={}, noWait=False):
        """Executes an SQL statement
        Args:
            stmt: a statement in immudb SQL dialect.
            params: a dictionary of parameters to replace in the statement
            noWait: whether to wait for indexing. Set to True for fast inserts.

        Returns:
            An object with two lists: ctxs and dtxs, including transaction
            metadata for both the catalog and the data store.

            Each element of both lists contains an object with the Transaction ID
            (id), timestamp (ts), and number of entries (nentries).
        """

        return sqlexec.call(self.__stub, self.__rs, stmt, params, noWait)

    def sqlQuery(self, query, params={}, columnNameMode=constants.COLUMN_NAME_MODE_NONE):
        """Queries the database using SQL
        Args:
            query: a query in immudb SQL dialect.
            params: a dictionary of parameters to replace in the query

        Returns:
            A list of table names. For example:

            ['table1', 'table2']
        """
        return sqlquery.call(self.__stub, self.__rs, query, params, columnNameMode)

    def listTables(self):
        """List all tables in the current database

        Returns:
            A list of table names. For example:

            ['table1', 'table2']
        """
        return listtables.call(self.__stub, self.__rs)

    def describeTable(self, table):
        return sqldescribe.call(self.__stub, self.__rs, table)

    # Not implemented: verifyRow

# deprecated
    def databaseCreate(self, dbName: bytes):
        warnings.warn("Call to deprecated databaseCreate. Use createDatabase instead",
                      category=DeprecationWarning,
                      stacklevel=2
                      )
        return self.createDatabase(dbName)

    def safeGet(self, key: bytes):  # deprecated
        warnings.warn("Call to deprecated safeGet. Use verifiedGet instead",
                      category=DeprecationWarning,
                      stacklevel=2
                      )
        return verifiedGet.call(self.__stub, self.__rs, key, verifying_key=self.__vk)

    def databaseUse(self, dbName: bytes):  # deprecated
        warnings.warn("Call to deprecated databaseUse. Use useDatabase instead",
                      category=DeprecationWarning,
                      stacklevel=2
                      )
        return self.useDatabase(dbName)

    def safeSet(self, key: bytes, value: bytes):  # deprecated
        warnings.warn("Call to deprecated safeSet. Use verifiedSet instead",
                      category=DeprecationWarning,
                      stacklevel=2
                      )
        return verifiedSet.call(self.__stub, self.__rs, key, value)


# immudb-py only

    def getAllValues(self, keys: list):  # immudb-py only
        resp = batchGet.call(self.__stub, self.__rs, keys)
        return resp

    def getValue(self, key: bytes):  # immudb-py only
        ret = get.call(self.__stub, self.__rs, key)
        if ret is None:
            return None
        return ret.value



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.embedded import store, ahtree
from immudb.constants import *
from immudb.exceptions import ErrUnsupportedTxVersion
import hashlib
import struct


def VerifyInclusion(proof, digest: bytes, root) -> bool:
    if proof == None:
        return False
    leaf = LEAF_PREFIX+digest
    calcRoot = hashlib.sha256(leaf).digest()
    i = proof.leaf
    r = proof.width-1
    for t in proof.terms:
        b = NODE_PREFIX
        if i % 2 == 0 and i != r:
            b = b+calcRoot+t
        else:
            b = b+t+calcRoot
        calcRoot = hashlib.sha256(b).digest()
        i = i//2
        r = r//2
    return i == r and root == calcRoot


def VerifyLinearProof(proof, sourceTxID: int, targetTxID: int, sourceAlh: bytes, targetAlh: bytes) -> bool:
    if proof == None or proof.sourceTxID != sourceTxID or proof.targetTxID != targetTxID:
        return False

    if (proof.sourceTxID == 0 or proof.sourceTxID > proof.targetTxID or
            len(proof.terms) == 0 or sourceAlh != proof.terms[0]):
        return False

    calculatedAlh = proof.terms[0]
    for i in range(1, len(proof.terms)):
        bs = struct.pack(">Q", proof.sourceTxID+i)+calculatedAlh+proof.terms[i]
        calculatedAlh = hashlib.sha256(bs).digest()

    return targetAlh == calculatedAlh


def VerifyDualProof(proof, sourceTxID, targetTxID, sourceAlh, targetAlh):
    if (proof == None or
        proof.sourceTxHeader == None or
        proof.targetTxHeader == None or
        proof.sourceTxHeader.iD != sourceTxID or
            proof.targetTxHeader.iD != targetTxID):
        return False
    if proof.sourceTxHeader.iD == 0 or proof.sourceTxHeader.iD > proof.targetTxHeader.iD:
        return False
    if sourceAlh != proof.sourceTxHeader.Alh():
        return False
    if targetAlh != proof.targetTxHeader.Alh():
        return False
    if sourceTxID < proof.targetTxHeader.blTxID and ahtree.VerifyInclusion(
            proof.inclusionProof,
            sourceTxID,
            proof.targetTxHeader.blTxID,
            leafFor(sourceAlh),
            proof.targetTxHeader.blRoot) == False:
        return False
    if proof.sourceTxHeader.blTxID > 0 and ahtree.VerifyConsistency(
            proof.consistencyProof,
            proof.sourceTxHeader.blTxID,
            proof.targetTxHeader.blTxID,
            proof.sourceTxHeader.blRoot,
            proof.targetTxHeader.blRoot) == False:
        return False
    if proof.targetTxHeader.blTxID > 0 and ahtree.VerifyLastInclusion(
            proof.lastInclusionProof,
            proof.targetTxHeader.blTxID,
            leafFor(proof.targetBlTxAlh),
            proof.targetTxHeader.blRoot) == False:
        return False
    if sourceTxID < proof.targetTxHeader.blTxID:
        ret = VerifyLinearProof(
            proof.linearProof, proof.targetTxHeader.blTxID, targetTxID, proof.targetBlTxAlh, targetAlh)
    else:
        ret = VerifyLinearProof(
            proof.linearProof, sourceTxID, targetTxID, sourceAlh, targetAlh)
    return ret


def EntrySpecDigestFor(version: int):
    if version == 0:
        return EntrySpecDigest_v0
    elif version == 1:
        return EntrySpecDigest_v1
    else:
        raise ErrUnsupportedTxVersion


def EntrySpecDigest_v0(kv: store.EntrySpec) -> bytes:
    md = hashlib.sha256()
    md.update(kv.key)
    valmd = hashlib.sha256()
    valmd.update(kv.value)
    md.update(valmd.digest())
    return md.digest()


def EntrySpecDigest_v1(kv: store.EntrySpec) -> bytes:
    mdbs = b''
    if kv.metadata != None:
        mdbs = kv.metadata.Bytes()
    mdLen = len(mdbs)
    kLen = len(kv.key)
    b = b''
    b = b+mdLen.to_bytes(2, 'big')
    b = b+mdbs
    b = b+kLen.to_bytes(2, 'big')
    b = b+kv.key

    md = hashlib.sha256()
    md.update(b)
    valmd = hashlib.sha256()
    valmd.update(kv.value)
    md.update(valmd.digest())
    return md.digest()


def leafFor(d: bytes) -> bytes:
    b = LEAF_PREFIX+d
    return hashlib.sha256(b).digest()



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from immudb.datatypes import ColumnDescription
from immudb.typeconv import sqlvalue_to_py
from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, table):
    res = service.DescribeTable(schema_pb2.Table(tableName=table.encode()))
    result = []
    for row in res.rows:
        result.append(
            ColumnDescription(
                name=sqlvalue_to_py(row.values[0]),
                type=sqlvalue_to_py(row.values[1]),
                nullable=sqlvalue_to_py(row.values[2]),
                index=sqlvalue_to_py(row.values[3]),
                autoincrement=sqlvalue_to_py(row.values[4]),
                unique=sqlvalue_to_py(row.values[5]),
            )
        )
    return result



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.constants import *
from immudb.printable import printable
import immudb.embedded.store as store


class EntrySpec(printable):

    def __init__(self, key: bytes, md: store.KVMetadata, value: bytes):
        self.key = key
        self.metadata = md
        self.value = value



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.constants import *
import hashlib


def VerifyInclusion(iproof: list, i: int, j: int, iLeaf: bytes, jRoot: bytes) -> bool:
    if i > j or i == 0 or i < j and len(iproof) == 0:
        return False
    ciRoot = EvalInclusion(iproof, i, j, iLeaf)
    return jRoot == ciRoot


def EvalInclusion(iproof: list, i: int, j: int, iLeaf: bytes) -> bool:
    i1 = i - 1
    j1 = j - 1
    ciRoot = iLeaf
    for h in iproof:
        if i1 % 2 == 0 and i1 != j1:
            b = NODE_PREFIX+ciRoot+h
        else:
            b = NODE_PREFIX+h+ciRoot
        ciRoot = hashlib.sha256(b).digest()
        i1 = i1 >> 1
        j1 = j1 >> 1
    return ciRoot


def VerifyConsistency(cproof: list, i: int, j: int, iRoot: bytes, jRoot: bytes) -> bool:
    if i > j or i == 0 or (i < j and len(cproof) == 0):
        return False
    if i == j and len(cproof) == 0:
        return iRoot == jRoot
    ciRoot, cjRoot = EvalConsistency(cproof, i, j)
    return iRoot == ciRoot and jRoot == cjRoot


def EvalConsistency(cproof: list, i: int, j: int):
    fn = i - 1
    sn = j - 1
    while fn % 2 == 1:
        fn = fn >> 1
        sn = sn >> 1
    ciRoot, cjRoot = cproof[0], cproof[0]
    for h in cproof[1:]:
        if fn % 2 == 1 or fn == sn:
            b = NODE_PREFIX+h+ciRoot
            ciRoot = hashlib.sha256(b).digest()
            b = NODE_PREFIX+h+cjRoot
            cjRoot = hashlib.sha256(b).digest()
            while fn % 2 == 0 and fn != 0:
                fn = fn >> 1
                sn = sn >> 1
        else:
            b = NODE_PREFIX+cjRoot+h
            cjRoot = hashlib.sha256(b).digest()
        fn = fn >> 1
        sn = sn >> 1
    return ciRoot, cjRoot


def VerifyLastInclusion(iproof: list, i: int, leaf: bytes, root: bytes) -> bool:
    if i == 0:
        return False
    return root == EvalLastInclusion(iproof, i, leaf)


def EvalLastInclusion(iproof: list, i: int, leaf: bytes) -> bytes:
    i1 = i - 1
    root = leaf
    for h in iproof:
        b = NODE_PREFIX+h+root
        root = hashlib.sha256(b).digest()
        i1 >>= 1
    return root



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb.typeconv import py_to_sqlvalue
from immudb.typeconv import sqlvalue_to_py
from immudb import constants
from immudb.exceptions import ErrPySDKInvalidColumnMode


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, query, params, columnNameMode):
    return _call_with_executor(query, params, columnNameMode, service.SQLQuery)


def _call_with_executor(query, params, columnNameMode, executor):
    paramsObj = []
    for key, value in params.items():
        paramsObj.append(schema_pb2.NamedParam(
            name=key, value=py_to_sqlvalue(value)))

    request = schema_pb2.SQLQueryRequest(
        sql=query,
        params=paramsObj)

    resp = executor(request)
    result = []

    columnNames = getColumnNames(resp, columnNameMode)

    for row in resp.rows:
        if columnNameMode == constants.COLUMN_NAME_MODE_NONE:
            result.append(tuple([sqlvalue_to_py(i) for i in row.values]))
        else:
            result.append(
                dict(zip(columnNames, tuple([sqlvalue_to_py(i) for i in row.values]))))
    return result


def getColumnNames(resp, columnNameMode):
    columnNames = []
    if columnNameMode:
        for column in resp.columns:
            if columnNameMode == constants.COLUMN_NAME_MODE_FIELD:
                columnNames.append(column.name.strip("()").split(".")[2])
            elif columnNameMode == constants.COLUMN_NAME_MODE_TABLE:
                columnNames.append(column.name.strip("()").split(".", 1)[1])
            elif columnNameMode == constants.COLUMN_NAME_MODE_DATABASE:
                columnNames.append(column.name.strip("()"))
            elif columnNameMode == constants.COLUMN_NAME_MODE_FULL:
                columnNames.append(column.name)
            else:
                raise ErrPySDKInvalidColumnMode
    return columnNames



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb import constants
from immudb.grpc import schema_pb2, schema_pb2_grpc
from google.protobuf import empty_pb2 as g_empty
import pickle
import os.path
import hashlib
import ecdsa
import ecdsa.util
import struct
from dataclasses import dataclass

_statefile = constants.ROOT_CACHE_PATH


@dataclass
class State(object):
    db: str
    txId: int
    txHash: bytes
    publicKey: bytes
    signature: bytes

    @staticmethod
    def FromGrpc(grpcState):
        rs = State(
            db=grpcState.db,
            txId=grpcState.txId,
            txHash=grpcState.txHash,
            publicKey=grpcState.signature.publicKey,
            signature=grpcState.signature.signature,
        )
        return rs

    def Hash(self):
        b = struct.pack(">I", len(self.db))+self.db.encode('utf8')
        b += struct.pack(">Q", self.txId)+self.txHash
        return b

    def Verify(self, verifying_key):
        verifying_key.verify(self.signature, self.Hash(
        ), hashlib.sha256, sigdecode=ecdsa.util.sigdecode_der)


# Sample reference implementation, with no state persistance on disk.
# Can be used when instancing more that one client, in order to
# avoid concurrency issues
class RootService:
    def __init__(self):
        self.__dbname = None
        self.__cache = None
        self.__service = None

    def init(self, dbname: str, service: schema_pb2_grpc.ImmuServiceStub):
        self.__dbname = dbname
        self.__service = service
        state = self.__service.CurrentState(g_empty.Empty())
        self.__cache = state

    def get(self) -> State:
        if self.__cache == None:
            self.__cache = self.__service.CurrentState(g_empty.Empty())
        return self.__cache

    def set(self, root: State):
        self.__cache = root

# Sample implementation of persistent state. A state file is created, with
# a dictionary of immudb states, one per database. State file name can be
# set at object creation, will use a default one if name is not set.
# Note: this implementation is not working correctly if more than one
# instance is using the same state file.
# Note: this implementation is not thread/process safe.


class PersistentRootService(RootService):
    def __init__(self, filename: str = None):
        self.__cache = None
        if filename != None:
            self.__filename = filename
        else:
            self.__filename = os.path.join(os.path.expanduser("~"), _statefile)

    def init(self, dbname: str, service: schema_pb2_grpc.ImmuServiceStub):
        self.__dbname = dbname
        self.__service = service
        self.__cache = None
        try:
            with open(self.__filename, "rb") as f:
                states = pickle.load(f)
                if self.__dbname in states:
                    self.__cache = states[self.__dbname]
                    # IMPROVEMENT: we could check here, if state is valid.
        except FileNotFoundError:
            pass
        except Exception as e:
            print("Warning:", e)
        if self.__cache == None:
            self.__cache = self.__service.CurrentState(g_empty.Empty())

    def get(self) -> State:
        if self.__cache == None:
            # should never occour this
            self.__cache = self.__service.CurrentState(g_empty.Empty())
        return self.__cache

    def set(self, root: State):
        self.__cache = root
        states = {}
        try:
            with open(self.__filename, "rb") as f:
                states = pickle.load(f)
        except Exception as e:
            print("Warning:", e)
        states[self.__dbname] = self.__cache
        with open(self.__filename, "wb") as f:
            pickle.dump(states, f)



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService, State
from immudb.exceptions import ErrCorruptedData
from immudb import datatypes
from immudb.embedded import store
import immudb.database as database
import immudb.schema as schema


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, zset: bytes, score: float, key: bytes, atTx: int = 0, verifying_key=None):
    state = rs.get()
    request = schema_pb2.VerifiableZAddRequest(
        zAddRequest=schema_pb2.ZAddRequest(
            set=zset,
            score=score,
            key=key,
            atTx=atTx,
        ),
        proveSinceTx=state.txId
    )
    vtx = service.VerifiableZAdd(request)
    if vtx.tx.header.nentries != 1:
        raise ErrCorruptedData
    tx = schema.TxFromProto(vtx.tx)
    entrySpecDigest = store.EntrySpecDigestFor(tx.header.version)
    ekv = database.EncodeZAdd(zset, score, key, atTx)
    inclusionProof = tx.Proof(ekv.key)
    verifies = store.VerifyInclusion(
        inclusionProof, entrySpecDigest(ekv), tx.header.eh)
    if not verifies:
        raise ErrCorruptedData
    if tx.header.eh != schema.DigestFromProto(vtx.dualProof.targetTxHeader.eH):
        raise ErrCorruptedData

    sourceID = state.txId
    sourceAlh = schema.DigestFromProto(state.txHash)
    targetID = tx.header.iD
    targetAlh = tx.header.Alh()
    if state.txId > 0:
        verifies = store.VerifyDualProof(
            schema.DualProofFromProto(vtx.dualProof),
            sourceID,
            targetID,
            sourceAlh,
            targetAlh,
        )
        if not verifies:
            raise ErrCorruptedData

    newstate = State(
        db=state.db,
        txId=targetID,
        txHash=targetAlh,
        publicKey=vtx.signature.publicKey,
        signature=vtx.signature.signature,
    )
    if verifying_key != None:
        newstate.Verify(verifying_key)
    rs.set(newstate)
    return datatypes.SetResponse(
        id=vtx.tx.header.id,
        verified=True,
    )



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.exceptions import ErrCorruptedData

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService, State
from immudb import datatypes
from immudb.embedded import store
import immudb.database as database
import immudb.schema as schema
from immudb.typeconv import MetadataToProto

#import base64


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, key: bytes, value: bytes, verifying_key=None, metadata=None):
    schemaMetadata = MetadataToProto(metadata)
    state = rs.get()
    # print(base64.b64encode(state.SerializeToString()))
    kv = schema_pb2.KeyValue(key=key, value=value, metadata=schemaMetadata)
    rawRequest = schema_pb2.VerifiableSetRequest(
        setRequest=schema_pb2.SetRequest(KVs=[kv]),
        proveSinceTx=state.txId,
    )
    verifiableTx = service.VerifiableSet(rawRequest)
    # print(base64.b64encode(verifiableTx.SerializeToString()))
    if verifiableTx.tx.header.nentries != 1 or len(verifiableTx.tx.entries) != 1:
        raise ErrCorruptedData
    tx = schema.TxFromProto(verifiableTx.tx)
    entrySpecDigest = store.EntrySpecDigestFor(tx.header.version)
    inclusionProof = tx.Proof(database.EncodeKey(key))
    md = tx.entries[0].metadata()

    if md != None and md.Deleted():
        raise ErrCorruptedData

    e = database.EncodeEntrySpec(key, md, value)

    verifies = store.VerifyInclusion(
        inclusionProof, entrySpecDigest(e), tx.header.eh)
    if not verifies:
        raise ErrCorruptedData
    if tx.header.eh != schema.DigestFromProto(verifiableTx.dualProof.targetTxHeader.eH):
        raise ErrCorruptedData
    sourceID = state.txId
    sourceAlh = schema.DigestFromProto(state.txHash)
    targetID = tx.header.iD
    targetAlh = tx.header.Alh()

    if state.txId > 0:
        verifies = store.VerifyDualProof(
            schema.DualProofFromProto(verifiableTx.dualProof),
            sourceID,
            targetID,
            sourceAlh,
            targetAlh,
        )
        if not verifies:
            raise ErrCorruptedData

    newstate = State(
        db=state.db,
        txId=targetID,
        txHash=targetAlh,
        publicKey=verifiableTx.signature.publicKey,
        signature=verifiableTx.signature.signature,
    )
    if verifying_key != None:
        newstate.Verify(verifying_key)
    rs.set(newstate)
    return datatypes.SetResponse(
        id=verifiableTx.tx.header.id,
        verified=verifies,
    )



class printable:
    def __repr__(self) -> str:
        ret = "class {}\n".format(type(self).__name__)
        for k in self.__dict__:
            v = self.__dict__[k]
            if type(v) == bytes:
                ret += "\t{}: {}\n".format(k, list(v))
            elif hasattr(v, '__getitem__') and type(v[0]) == bytes:
                ret += "\t{}:\n".format(k)
                for t in v:
                    ret += "\t- {}\n".format(list(t))
            else:
                ret += "\t{}: {}\n".format(k, v)
        return ret



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.embedded import store, htree
from immudb.printable import printable
import hashlib
from immudb.exceptions import ErrCorruptedData, ErrKeyNotFound, ErrMetadataUnsupported
import sys


def NewTxWithEntries(header: "TxHeader", entries: list) -> "Tx":

    _htree = htree.HTree(len(entries))

    tx = Tx()
    tx.header = header
    tx.entries = entries
    tx.htree = _htree
    return tx


class TxHeader(printable):
    def __init__(self):
        self.iD = None
        self.ts = None
        self.blTxID = None
        self.blRoot = None
        self.prevAlh = None

        self.version = None
        self.metadata = store.TxMetadata()

        self.nentries = None
        self.eh = None

    def innerHash(self) -> bytes:
        md = hashlib.sha256()
        md.update(self.ts.to_bytes(8, 'big'))
        md.update(self.version.to_bytes(2, 'big'))
        if self.version == 0:
            md.update(self.nentries.to_bytes(2, 'big'))
        elif self.version == 1:
            mdbs = b''
            if self.metadata != None:
                mdbs = self.metadata.Bytes()
                if mdbs == None:
                    mdbs = b''
            md.update(len(mdbs).to_bytes(2, 'big'))
            md.update(mdbs)
            md.update(self.nentries.to_bytes(4, 'big'))
        else:
            sys.exit("missing tx hash calculation method for version %d" %
                     self.version)
        md.update(self.eh)
        md.update(self.blTxID.to_bytes(8, 'big'))
        md.update(self.blRoot)
        return md.digest()

    def Alh(self) -> bytes:
        md = hashlib.sha256()
        md.update(self.iD.to_bytes(8, 'big'))
        md.update(self.prevAlh)
        md.update(self.innerHash())
        return md.digest()


class Tx(printable):
    def __init__(self):
        self.header = None
        self.entries = None
        self.htree = None

    def TxEntryDigest(self):
        if self.header.version == 0:
            return TxEntryDigest_v1_1
        elif self.header.version == 1:
            return TxEntryDigest_v1_2
        else:
            raise ErrCorruptedData

    def BuildHashTree(self):
        digests = []
        txEntryDigest = self.TxEntryDigest()
        for e in self.entries:
            digests.append(txEntryDigest(e))
        self.htree.BuildWith(digests)
        root = self.htree.root
        self.header.eh = root

    def IndexOf(self, key: bytes) -> int:
        kindex = None
        for k, v in enumerate(self.entries):
            if v.key() == key:
                kindex = k
                break
        if kindex == None:
            raise ErrKeyNotFound
        return k

    def Proof(self, key: bytes):
        kindex = self.IndexOf(key)
        return self.htree.InclusionProof(kindex)


class TxEntry(printable):
    def __init__(self, key: bytes, md: store.KVMetadata, vLen: int, hVal: bytes, vOff: int):
        self.k = key
        self.kLen = len(key)
        self.md = md
        self.vLen = vLen
        self.hVal = hVal
        self.vOff = vOff

    def key(self) -> bytes:
        return self.k

    def metadata(self) -> store.KVMetadata:
        return self.md


def TxEntryDigest_v1_1(e: TxEntry) -> bytes:
    if len(e.md.Bytes()) > 0:  # n.b. contrary to Golang, in Pyhton e.md will not be None
        raise ErrMetadataUnsupported
    md = hashlib.sha256()
    md.update(e.k)
    md.update(e.hVal)
    return md.digest()


def TxEntryDigest_v1_2(e: TxEntry) -> bytes:
    mdbs = b''
    if e.md != None:
        mdbs = e.md.Bytes()
    mdLen = len(mdbs)
    b = b''
    b = b+mdLen.to_bytes(2, 'big')
    b = b+mdbs
    b = b+e.kLen.to_bytes(2, 'big')
    b = b+e.k
    md = hashlib.sha256()
    md.update(b)
    md.update(e.hVal)
    return md.digest()



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb.exceptions import ErrCorruptedData
from immudb import datatypes


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, zset: bytes, score: float, key: bytes, atTx: int = 0):
    request = schema_pb2.ZAddRequest(
        set=zset,
        score=score,
        key=key,
        atTx=atTx,
        boundRef=atTx > 0,
    )
    msg = service.ZAdd(request)
    if msg.nentries != 1:
        raise ErrCorruptedData
    return datatypes.SetResponse(
        id=msg.id,
        verified=False,
    )



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.datatypes import DeleteKeysRequest

from immudb.grpc import schema_pb2_grpc


def call(service: schema_pb2_grpc.ImmuServiceStub, req: DeleteKeysRequest):
    request = schema_pb2_grpc.schema__pb2.DeleteKeysRequest(
        keys=req.keys,
        sinceTx=req.sinceTx,
        noWait=req.noWait
    )
    msg = service.Delete(request)
    return msg



from immudb.grpc import schema_pb2
from pprint import pformat
from datetime import datetime, timezone
from immudb.embedded.store import KVMetadata


def py_to_sqlvalue(value):
    sqlValue = None
    typ = type(value)
    if value is None:
        sqlValue = schema_pb2.SQLValue(null=True)
    elif typ is int:
        sqlValue = schema_pb2.SQLValue(n=value)
    elif typ is bool:
        sqlValue = schema_pb2.SQLValue(b=value)
    elif typ is str:
        sqlValue = schema_pb2.SQLValue(s=value)
    elif typ in (bytes, bytearray):
        sqlValue = schema_pb2.SQLValue(bs=value)
    elif typ is datetime:
        sqlValue = schema_pb2.SQLValue(
            ts=int(value.timestamp()*1e6))
    else:
        raise TypeError("Type not supported: {}".format(
            value.__class__.__name__))
    return sqlValue


def sqlvalue_to_py(sqlValue):
    if sqlValue.HasField("n"):
        return sqlValue.n
    elif sqlValue.HasField("b"):
        return bool(sqlValue.b)
    elif sqlValue.HasField("bs"):
        return sqlValue.bs
    elif sqlValue.HasField("s"):
        return sqlValue.s
    elif sqlValue.HasField("ts"):
        return datetime.fromtimestamp(sqlValue.ts/1e6, timezone.utc)
    elif sqlValue.HasField("null"):
        return None
    else:
        raise TypeError("Type not supported: {}".format(
            sqlValue.WhichOneof("value")))


def MetadataToProto(metadata: KVMetadata):
    schemaMetadata = None
    if metadata:
        schemaMetadata = schema_pb2.KVMetadata()
        if metadata.Deleted():
            schemaMetadata.deleted = True
        if metadata.IsExpirable():
            schemaMetadata.expiration.expiresAt = int(
                metadata.ExpirationTime().timestamp())
        if metadata.NonIndexable():
            schemaMetadata.nonIndexable = True
    return schemaMetadata



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ErrMaxWidthExceeded(Exception):
    pass


class ErrIllegalArguments(Exception):
    pass


class ErrUnsupportedTxVersion(Exception):
    pass


class ErrNonExpirable(Exception):
    pass


class ErrCorruptedData(Exception):
    pass


# Compatibility with older SDK
VerificationException = ErrCorruptedData


class ErrReadOnly(Exception):
    pass


class ErrKeyNotFound(Exception):
    pass


class ErrMetadataUnsupported(Exception):
    pass


class ErrPySDKInvalidColumnMode(Exception):
    pass



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService


@dataclass
class batchElement:
    tx: int
    key: bytes
    value: bytes


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, keys: list):
    request = schema_pb2.KeyListRequest(
        keys=keys  # [schema_pb2_grpc.schema__pb2.Key(key=k) for k in keys]
    )
    msg = service.GetAll(request)
    ret = {}
    for i in msg.entries:
        element = batchElement(
            tx=i.tx,
            key=i.key,
            value=i.value
        )
        ret[i.key] = element
    return ret



#!/usr/bin/env python3
# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='immudb-py',
      version='1.3.1',
      license="Apache License Version 2.0",
      description='Python SDK for Immudb',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Codenotary',
      url='https://github.com/codenotary/immudb-py',
      # download_url='',
      packages=['immudb', 'immudb.database', 'immudb.embedded',
                'immudb.embedded.ahtree', 'immudb.embedded.htree', 'immudb.embedded.store',
                'immudb.grpc', 'immudb.handler', 'immudb.schema'],
      keywords=['immudb', 'immutable'],
      install_requires=[
          'grpcio>=1.31.0',
          'dataclasses>=0.6',
          'protobuf>=3.13.0,<4.0.0',
          'google-api>=0.1.12',
          'google-api-core>=1.22.1',
          'ecdsa>=0.16.1'
      ],
      classifiers=[
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Build Tools',
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
      ],
      python_requires='>=3.6',
      )



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.constants import *
import immudb.embedded.store as store
import struct


def WrapWithPrefix(b: bytes, prefix: bytes) -> bytes:
    return prefix+b


def EncodeKey(key: bytes):
    return WrapWithPrefix(key, SET_KEY_PREFIX)


def EncodeEntrySpec(key: bytes, md: store.KVMetadata, value: bytes):
    es = store.EntrySpec(
        key=WrapWithPrefix(key, SET_KEY_PREFIX),
        md=md,
        value=WrapWithPrefix(value, PLAIN_VALUE_PREFIX)
    )
    return es


def EncodeReference(key: bytes, md: store.KVMetadata, referencedKey: bytes, atTx: int):
    es = store.EntrySpec(key=WrapWithPrefix(key, SET_KEY_PREFIX),
                         md=md,
                         value=WrapReferenceValueAt(
        WrapWithPrefix(referencedKey, SET_KEY_PREFIX), atTx)
    )
    return es


def WrapReferenceValueAt(key: bytes, atTx: int) -> bytes:
    refVal = REFERENCE_VALUE_PREFIX+atTx.to_bytes(8, 'big')+key
    return refVal


def EncodeZAdd(zset: bytes, score: float, key: bytes, attx: int):
    es = store.EntrySpec(key=WrapZAddReferenceAt(zset, score, key, attx),
                         md=None,
                         value=b''
                         )
    return es


def WrapZAddReferenceAt(zset: bytes, score: float, key: bytes, attx: int) -> bytes:
    ekey = SET_KEY_PREFIX+key
    zkey = SORTED_KEY_PREFIX
    zkey += struct.pack(">Q", len(zset))+zset
    zkey += struct.pack(">d", score)
    zkey += struct.pack(">Q", len(ekey))+ekey
    zkey += struct.pack(">Q", attx)
    return zkey



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.constants import *
from immudb.embedded import store, htree
import copy
from immudb.grpc.schema_pb2 import KVMetadata as grpc_KVMetadata
from immudb.grpc.schema_pb2 import TxMetadata as grpc_TxMetadata
from immudb.grpc.schema_pb2 import DualProof as grpc_DualProof
from immudb.grpc.schema_pb2 import TxHeader as grpc_TxHeader
from immudb.grpc.schema_pb2 import LinearProof as grpc_LinearProof
from immudb.grpc.schema_pb2 import InclusionProof as grpc_InclusionProof
from datetime import datetime

import hashlib


def TxFromProto(stx) -> store.Tx():
    header = store.TxHeader()
    header.iD = stx.header.id
    header.ts = stx.header.ts
    header.blTxID = stx.header.blTxId
    header.blRoot = DigestFromProto(stx.header.blRoot)
    header.prevAlh = DigestFromProto(stx.header.prevAlh)

    header.version = int(stx.header.version)
    header.metadata = TxMetadataFromProto(stx.header.metadata)

    header.nentries = int(stx.header.nentries)
    header.eh = DigestFromProto(stx.header.eH)

    entries = []
    for e in stx.entries:
        entries.append(
            store.TxEntry(
                e.key,
                KVMetadataFromProto(e.metadata),
                int(e.vLen),
                DigestFromProto(e.hValue),
                0)
        )

    tx = store.NewTxWithEntries(header, entries)

    tx.BuildHashTree()

    return tx


def KVMetadataFromProto(md: grpc_KVMetadata) -> store.KVMetadata:
    if md == None:
        return None
    kvmd = store.KVMetadata()
    kvmd.AsDeleted(md.deleted)

    if md.HasField("expiration"):
        kvmd.ExpiresAt(datetime.utcfromtimestamp(md.expiration.expiresAt))

    kvmd.AsNonIndexable(md.nonIndexable)

    return kvmd


def InclusionProofFromProto(iproof: grpc_InclusionProof) -> htree.InclusionProof:
    ip = htree.InclusionProof()
    ip.leaf = int(iproof.leaf)
    ip.width = int(iproof.width)
    ip.terms = DigestsFromProto(iproof.terms)
    return ip


def DualProofFromProto(dproof: grpc_DualProof) -> store.DualProof:
    dp = store.DualProof()
    dp.sourceTxHeader = TxHeaderFromProto(dproof.sourceTxHeader)
    dp.targetTxHeader = TxHeaderFromProto(dproof.targetTxHeader)
    dp.inclusionProof = DigestsFromProto(dproof.inclusionProof)
    dp.consistencyProof = DigestsFromProto(dproof.consistencyProof)
    dp.targetBlTxAlh = DigestFromProto(dproof.targetBlTxAlh)
    dp.lastInclusionProof = DigestsFromProto(dproof.lastInclusionProof)
    dp.linearProof = LinearProofFromProto(dproof.linearProof)
    return dp


def TxHeaderFromProto(hdr: grpc_TxHeader) -> store.TxHeader:
    txh = store.TxHeader()
    txh.iD = hdr.id
    txh.prevAlh = DigestFromProto(hdr.prevAlh)
    txh.ts = hdr.ts
    txh.version = int(hdr.version)
    txh.metadata = TxMetadataFromProto(hdr.metadata)
    txh.nentries = int(hdr.nentries)
    txh.eh = DigestFromProto(hdr.eH)
    txh.blTxID = hdr.blTxId
    txh.blRoot = DigestFromProto(hdr.blRoot)
    return txh


def TxMetadataFromProto(md: grpc_TxMetadata) -> store.TxMetadata:
    if md == None:
        return None
    return store.TxMetadata()


def LinearProofFromProto(lproof: grpc_LinearProof) -> store.LinearProof:
    lp = store.LinearProof()
    lp.sourceTxID = lproof.sourceTxId
    lp.targetTxID = lproof.TargetTxId
    lp.terms = DigestsFromProto(lproof.terms)
    return lp


def DigestFromProto(slicedDigest: bytes) -> bytes:
    d = copy.copy(slicedDigest[:hashlib.sha256().digest_size])
    return d


def DigestsFromProto(slicedTerms: list) -> list:
    d = [copy.copy(i) for i in slicedTerms]
    return d



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService
from immudb.typeconv import py_to_sqlvalue


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, stmt, params, noWait):
    return _call_with_executor(stmt, params, noWait, service.SQLExec)


def _call_with_executor(stmt, params, noWait, executor):
    paramsObj = []
    for key, value in params.items():
        paramsObj.append(schema_pb2.NamedParam(
            name=key, value=py_to_sqlvalue(value)))

    request = schema_pb2.SQLExecRequest(
        sql=stmt,
        params=paramsObj,
        noWait=noWait)

    result = executor(request)
    return result



# Copyright 2021 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb import datatypes
from immudb.grpc import schema_pb2, schema_pb2_grpc
from immudb.rootService import RootService


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, kv: dict):
    request = schema_pb2.SetRequest(
        KVs=[schema_pb2.KeyValue(key=key, value=value)
             for key, value in kv.items()]
    )
    msg = service.Set(request)
    return datatypes.SetResponse(
        id=msg.id,
        verified=False,
    )



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.embedded import store
from immudb.grpc import schema_pb2
from immudb.grpc import schema_pb2_grpc
from immudb.rootService import RootService, State
from immudb import datatypes
from immudb.exceptions import ErrCorruptedData
import immudb.database as database
import immudb.schema as schema


import sys


def call(service: schema_pb2_grpc.ImmuServiceStub, rs: RootService, requestkey: bytes, atTx: int = None, verifying_key=None, sinceTx: int = None, atRevision: int = None):
    state = rs.get()
    req = schema_pb2.VerifiableGetRequest(
        keyRequest=schema_pb2.KeyRequest(
            key=requestkey, atTx=atTx, sinceTx=sinceTx, atRevision=atRevision),
        proveSinceTx=state.txId
    )
    ventry = service.VerifiableGet(req)
    entrySpecDigest = store.EntrySpecDigestFor(
        int(ventry.verifiableTx.tx.header.version))
    inclusionProof = schema.InclusionProofFromProto(ventry.inclusionProof)
    dualProof = schema.DualProofFromProto(ventry.verifiableTx.dualProof)

    if not ventry.entry.HasField("referencedBy"):
        vTx = ventry.entry.tx
        e = database.EncodeEntrySpec(requestkey, schema.KVMetadataFromProto(
            ventry.entry.metadata), ventry.entry.value)
    else:
        ref = ventry.entry.referencedBy
        vTx = ref.tx
        e = database.EncodeReference(ref.key, schema.KVMetadataFromProto(
            ref.metadata), ventry.entry.key, ref.atTx)

    if state.txId <= vTx:
        eh = schema.DigestFromProto(
            ventry.verifiableTx.dualProof.targetTxHeader.eH)
        sourceid = state.txId
        sourcealh = schema.DigestFromProto(state.txHash)
        targetid = vTx
        targetalh = dualProof.targetTxHeader.Alh()
    else:
        eh = schema.DigestFromProto(
            ventry.verifiableTx.dualProof.sourceTxHeader.eH)
        sourceid = vTx
        sourcealh = dualProof.sourceTxHeader.Alh()
        targetid = state.txId
        targetalh = schema.DigestFromProto(state.txHash)

    verifies = store.VerifyInclusion(inclusionProof, entrySpecDigest(e), eh)
    if not verifies:
        raise ErrCorruptedData

    if state.txId > 0:
        verifies = store.VerifyDualProof(
            dualProof,
            sourceid,
            targetid,
            sourcealh,
            targetalh)
        if not verifies:
            raise ErrCorruptedData
    newstate = State(
        db=state.db,
        txId=targetid,
        txHash=targetalh,
        publicKey=ventry.verifiableTx.signature.publicKey,
        signature=ventry.verifiableTx.signature.signature,
    )
    if verifying_key != None:
        newstate.Verify(verifying_key)
    rs.set(newstate)
    if ventry.entry.HasField("referencedBy"):
        refkey = ventry.entry.referencedBy.key
    else:
        refkey = None

    return datatypes.SafeGetResponse(
        id=vTx,
        key=ventry.entry.key,
        value=ventry.entry.value,
        timestamp=ventry.verifiableTx.tx.header.ts,
        verified=verifies,
        refkey=refkey,
        revision=ventry.entry.revision
    )



# Copyright 2022 CodeNotary, Inc. All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from immudb.constants import *
from immudb.exceptions import ErrNonExpirable, ErrReadOnly
import datetime


DELETED_ATTR_CODE = 0
EXPIRES_AT_ATTR_CODE = 1
NON_INDEXABLE_ATTR_CODE = 2


class KVMetadata():
    def __init__(self):
        self.attributes = dict()
        self.readonly = False

    def AsDeleted(self, deleted: bool):
        if self.readonly:
            raise ErrReadOnly
        if not deleted:
            self.attributes.pop(DELETED_ATTR_CODE, None)
        else:
            self.attributes[DELETED_ATTR_CODE] = None

    def Deleted(self) -> bool:
        return DELETED_ATTR_CODE in self.attributes

    def ExpiresAt(self, expiresAt: datetime.datetime):
        if self.readonly:
            raise ErrReadOnly
        self.attributes[EXPIRES_AT_ATTR_CODE] = expiresAt

    def NonExpirable(self) -> bool:
        self.attributes.pop(EXPIRES_AT_ATTR_CODE, None)

    def IsExpirable(self) -> bool:
        return EXPIRES_AT_ATTR_CODE in self.attributes

    def ExpirationTime(self) -> datetime.datetime:
        if EXPIRES_AT_ATTR_CODE in self.attributes:
            return self.attributes[EXPIRES_AT_ATTR_CODE]
        else:
            raise ErrNonExpirable

    def AsNonIndexable(self, nonIndexable: bool):
        if self.readonly:
            raise ErrReadOnly
        if not nonIndexable:
            self.attributes.pop(NON_INDEXABLE_ATTR_CODE, None)
        else:
            self.attributes[NON_INDEXABLE_ATTR_CODE] = None

    def NonIndexable(self) -> bool:
        return NON_INDEXABLE_ATTR_CODE in self.attributes

    def Bytes(self):
        b = b''
        for attrCode in [DELETED_ATTR_CODE, EXPIRES_AT_ATTR_CODE, NON_INDEXABLE_ATTR_CODE]:
            if attrCode in self.attributes:
                b = b+attrCode.to_bytes(1, 'big')
                if attrCode == EXPIRES_AT_ATTR_CODE:
                    # The attribute is a datetime. Convert it to epoch integer big endian 64 bit
                    b = b + \
                        int(self.attributes[attrCode].replace(tzinfo=datetime.timezone.utc).timestamp()).to_bytes(
                            8, 'big')
        return b
